﻿//+------------------------------------------------------------------+
//|                                               IndicatorClass.mqh |
//|                    Copyright 2022, Manuel Alejandro Cercós Pérez |
//|                         https://www.mql5.com/en/users/alexcercos |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, Manuel Alejandro Cercós Pérez"
#property link      "https://www.mql5.com/en/users/alexcercos"

//+------------------------------------------------------------------+
//| Base class for plots and calculation buffers                     |
//+------------------------------------------------------------------+
class CIndicatorBufferBase
{
public:
   virtual void      SetBuffer(int &buffer, int &plot)=NULL;
   virtual void      SetAsSeries(bool set)=NULL;

   virtual void      Init() { }
   virtual void      DeInit() { }
   virtual void      Update(const int rates_total,
                            const int prev_calculated,
                            const datetime &time[],
                            const double &open[],
                            const double &high[],
                            const double &low[],
                            const double &close[],
                            const long &tick_volume[],
                            const long &volume[],
                            const int &spread[]) { }
};

//+------------------------------------------------------------------+
//| Calculations Buffer (with no plot)                               |
//+------------------------------------------------------------------+
class CIndicatorCalculations : public CIndicatorBufferBase
{
public:
   double            array[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicatorCalculations::SetBuffer(int &index, int &plot)
{
   SetIndexBuffer(index, array, INDICATOR_CALCULATIONS);

   index++;
//No plot is used
}

//+------------------------------------------------------------------+
void CIndicatorCalculations::SetAsSeries(bool set)
{
   ArraySetAsSeries(array, set);
}

//+------------------------------------------------------------------+
//| Base indicator plot class                                        |
//+------------------------------------------------------------------+
class CIndicatorPlotBase : public CIndicatorBufferBase
{
protected:

   int               indicator_plot;

   virtual void      SetDrawType()=NULL; //Implicit in each class

public:

   void              SetArrow(uchar arrow);
   void              SetArrowShift(int shift);
   void              SetDrawBegin(int begin);
   void              SetShowData(bool show);
   void              SetShift(int shift);
   void              SetLineStyle(ENUM_LINE_STYLE style);
   void              SetLineWidth(int width);
   void              SetColorIndexes(color &color_array[]);
   void              SetLineColor(color line_color);
   void              SetLineColor(color line_color, int index);
   void              SetEmptyValue(double empty);
   void              SetLabel(string label);

   int               GetInteger(ENUM_PLOT_PROPERTY_INTEGER property_id, int property_modifier=0);
};

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetArrow(uchar arrow)
{
   PlotIndexSetInteger(indicator_plot, PLOT_ARROW, arrow);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetArrowShift(int shift)
{
   PlotIndexSetInteger(indicator_plot, PLOT_ARROW_SHIFT, shift);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetDrawBegin(int begin)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_BEGIN, begin);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetShowData(bool show)
{
   PlotIndexSetInteger(indicator_plot, PLOT_SHOW_DATA, show);
}
//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetShift(int shift)
{
   PlotIndexSetInteger(indicator_plot, PLOT_SHIFT, shift);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetLineStyle(ENUM_LINE_STYLE style)
{
   PlotIndexSetInteger(indicator_plot, PLOT_LINE_STYLE, style);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetLineWidth(int width)
{
   PlotIndexSetInteger(indicator_plot, PLOT_LINE_WIDTH, width);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetColorIndexes(color &color_array[])
{
   int amount = ArraySize(color_array);

   PlotIndexSetInteger(indicator_plot, PLOT_COLOR_INDEXES, amount);

   for (int i=0; i<amount; i++)
      PlotIndexSetInteger(indicator_plot, PLOT_LINE_COLOR, i, color_array[i]);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetLineColor(color line_color)
{
   PlotIndexSetInteger(indicator_plot, PLOT_LINE_COLOR, line_color);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetLineColor(color line_color, int index)
{
   PlotIndexSetInteger(indicator_plot, PLOT_LINE_COLOR, index, line_color);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetEmptyValue(double empty)
{
   PlotIndexSetDouble(indicator_plot, PLOT_EMPTY_VALUE, empty);
}

//+------------------------------------------------------------------+
void CIndicatorPlotBase::SetLabel(string label)
{
   PlotIndexSetString(indicator_plot, PLOT_LABEL, label);
}

//+------------------------------------------------------------------+
int CIndicatorPlotBase::GetInteger(ENUM_PLOT_PROPERTY_INTEGER property_id, int property_modifier=0)
{
   return PlotIndexGetInteger(indicator_plot, property_id, property_modifier);
}

//+------------------------------------------------------------------+
//| Base for indicators with 1 Data Buffer                           |
//+------------------------------------------------------------------+
class CIndicator_1Data : public CIndicatorPlotBase
{
public:

   double            array[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicator_1Data::SetBuffer(int &buffer,int &plot)
{
   indicator_plot = plot;

   SetIndexBuffer(buffer, array, INDICATOR_DATA);
   SetDrawType();

   buffer++;
   plot++;
}

//+------------------------------------------------------------------+
void CIndicator_1Data::SetAsSeries(bool set)
{
   ArraySetAsSeries(array, set);
}

//+------------------------------------------------------------------+
//| Plot Line (1 data buffer)                                        |
//+------------------------------------------------------------------+
class CIndicatorPlotLine : public CIndicator_1Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotLine::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_LINE);
}

//+------------------------------------------------------------------+
//| Plot None (1 data buffer)                                        |
//+------------------------------------------------------------------+
class CIndicatorPlotNone : public CIndicator_1Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotNone::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_NONE);
}

//+------------------------------------------------------------------+
//| Plot Section (1 data buffer)                                     |
//+------------------------------------------------------------------+
class CIndicatorPlotSection : public CIndicator_1Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotSection::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_SECTION);
}

//+------------------------------------------------------------------+
//| Plot Histogram (1 data buffer)                                   |
//+------------------------------------------------------------------+
class CIndicatorPlotHistogram : public CIndicator_1Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotHistogram::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_HISTOGRAM);
}

//+------------------------------------------------------------------+
//| Plot Arrow (1 data buffer)                                       |
//+------------------------------------------------------------------+
class CIndicatorPlotArrow : public CIndicator_1Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotArrow::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_ARROW);
}

//+------------------------------------------------------------------+
//| Base for indicators with 2 Data Buffers                          |
//+------------------------------------------------------------------+
class CIndicator_2Data : public CIndicatorPlotBase
{
public:

   double            first_array[];
   double            second_array[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};


//+------------------------------------------------------------------+
void CIndicator_2Data::SetBuffer(int &buffer, int &plot)
{
   indicator_plot = plot;

   SetIndexBuffer(buffer, first_array, INDICATOR_DATA);
   SetIndexBuffer(buffer+1, second_array, INDICATOR_DATA);
   SetDrawType();

   buffer+=2;
   plot++;
}

//+------------------------------------------------------------------+
void CIndicator_2Data::SetAsSeries(bool set)
{
   ArraySetAsSeries(first_array, set);
   ArraySetAsSeries(second_array, set);
}

//+------------------------------------------------------------------+
//| Plot Histogram 2 (2 data buffers)                                |
//+------------------------------------------------------------------+
class CIndicatorPlotHistogram2 : public CIndicator_2Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotHistogram2::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_HISTOGRAM2);
}

//+------------------------------------------------------------------+
//| Plot Zig-Zag (2 data buffers)                                    |
//+------------------------------------------------------------------+
class CIndicatorPlotZigZag : public CIndicator_2Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotZigZag::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_ZIGZAG);
}

//+------------------------------------------------------------------+
//| Plot Filling (2 data buffers)                                    |
//+------------------------------------------------------------------+
class CIndicatorPlotFilling : public CIndicator_2Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotFilling::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_FILLING);
}

//+------------------------------------------------------------------+
//| Base for indicators with 4 Data Buffers                          |
//+------------------------------------------------------------------+
class CIndicator_4Data : public CIndicatorPlotBase
{
public:

   double            open_array[], high_array[], low_array[], close_array[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicator_4Data::SetBuffer(int &buffer,int &plot)
{
   indicator_plot = plot;

   SetIndexBuffer(buffer, open_array, INDICATOR_DATA);
   SetIndexBuffer(buffer+1, high_array, INDICATOR_DATA);
   SetIndexBuffer(buffer+2, low_array, INDICATOR_DATA);
   SetIndexBuffer(buffer+3, close_array, INDICATOR_DATA);
   SetDrawType();

   buffer+=4;
   plot++;
}

//+------------------------------------------------------------------+
void CIndicator_4Data::SetAsSeries(bool set)
{
   ArraySetAsSeries(open_array, set);
   ArraySetAsSeries(high_array, set);
   ArraySetAsSeries(low_array, set);
   ArraySetAsSeries(close_array, set);
}

//+------------------------------------------------------------------+
//| Plot candles (4 data buffers)                                    |
//+------------------------------------------------------------------+
class CIndicatorPlotCandles : public CIndicator_4Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotCandles::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_CANDLES);
}

//+------------------------------------------------------------------+
//| Plot Bars (4 data buffers)                                       |
//+------------------------------------------------------------------+
class CIndicatorPlotBars : public CIndicator_4Data
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotBars::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_BARS);
}

//+------------------------------------------------------------------+
//| Base for indicators with 1 Data Buffer & 1 Color Buffer          |
//+------------------------------------------------------------------+
class CIndicator_1Data1Color : public CIndicator_1Data
{
public:

   double            color_buffer[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicator_1Data1Color::SetBuffer(int &buffer, int &plot)
{
   CIndicator_1Data::SetBuffer(buffer, plot);

   SetIndexBuffer(buffer, color_buffer, INDICATOR_COLOR_INDEX);

   buffer++; //Add color buffer
}

//+------------------------------------------------------------------+
void CIndicator_1Data1Color::SetAsSeries(bool set)
{
   CIndicator_1Data::SetAsSeries(set);
   ArraySetAsSeries(color_buffer, set);
}

//+------------------------------------------------------------------+
//| Plot Color Line (1 data buffer, 1 color buffer)                  |
//+------------------------------------------------------------------+
class CIndicatorPlotColorLine : public CIndicator_1Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorLine::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_LINE);
}

//+------------------------------------------------------------------+
//| Plot Color Section (1 data buffer, 1 color buffer)               |
//+------------------------------------------------------------------+
class CIndicatorColorSection : public CIndicator_1Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorColorSection::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_SECTION);
}

//+------------------------------------------------------------------+
//| Plot Color Histogram (1 data buffer, 1 color buffer)             |
//+------------------------------------------------------------------+
class CIndicatorPlotColorHistogram : public CIndicator_1Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorHistogram::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_HISTOGRAM);
}

//+------------------------------------------------------------------+
//| Plot Color Arrow (1 data buffer, 1 color buffer)                 |
//+------------------------------------------------------------------+
class CIndicatorPlotColorArrow : public CIndicator_1Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorArrow::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_ARROW);
}

//+------------------------------------------------------------------+
//| Base for indicators with 2 Data Buffers & 1 Color Buffer         |
//+------------------------------------------------------------------+
class CIndicator_2Data1Color : public CIndicator_2Data
{
public:

   double            color_buffer[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicator_2Data1Color::SetBuffer(int &buffer, int &plot)
{
   CIndicator_2Data::SetBuffer(buffer, plot);

   SetIndexBuffer(buffer, color_buffer, INDICATOR_COLOR_INDEX);

   buffer++; //Add color buffer
}

//+------------------------------------------------------------------+
void CIndicator_2Data1Color::SetAsSeries(bool set)
{
   CIndicator_2Data::SetAsSeries(set);
   ArraySetAsSeries(color_buffer, set);
}

//+------------------------------------------------------------------+
//| Plot Color Histogram 2 (2 data buffers, 1 color buffer)          |
//+------------------------------------------------------------------+
class CIndicatorPlotColorHistogram2 : public CIndicator_2Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorHistogram2::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_HISTOGRAM2);
}

//+------------------------------------------------------------------+
//| Plot Color Zig Zag (2 data buffers, 1 color buffer)              |
//+------------------------------------------------------------------+
class CIndicatorPlotColorZigZag : public CIndicator_2Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorZigZag::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_ZIGZAG);
}

//+------------------------------------------------------------------+
//| Base for indicators with 2 Data Buffers & 1 Color Buffer         |
//+------------------------------------------------------------------+
class CIndicator_4Data1Color : public CIndicator_4Data
{
public:

   double            color_buffer[];

   virtual void      SetBuffer(int &buffer, int &plot);
   virtual void      SetAsSeries(bool set);
};

//+------------------------------------------------------------------+
void CIndicator_4Data1Color::SetBuffer(int &buffer, int &plot)
{
   CIndicator_4Data::SetBuffer(buffer, plot);

   SetIndexBuffer(buffer, color_buffer, INDICATOR_COLOR_INDEX);

   buffer++; //Add color buffer
}

//+------------------------------------------------------------------+
void CIndicator_4Data1Color::SetAsSeries(bool set)
{
   CIndicator_4Data::SetAsSeries(set);
   ArraySetAsSeries(color_buffer, set);
}

//+------------------------------------------------------------------+
//| Plot Color Candles (4 data buffers, 1 color buffer)              |
//+------------------------------------------------------------------+
class CIndicatorPlotColorCandles : public CIndicator_4Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorCandles::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_CANDLES);
}

//+------------------------------------------------------------------+
//| Plot Color Bars (4 data buffers, 1 color buffer)                 |
//+------------------------------------------------------------------+
class CIndicatorPlotColorBars : public CIndicator_4Data1Color
{
protected:

   virtual void      SetDrawType() final;
};

//+------------------------------------------------------------------+
void CIndicatorPlotColorBars::SetDrawType(void)
{
   PlotIndexSetInteger(indicator_plot, PLOT_DRAW_TYPE, DRAW_COLOR_BARS);
}
//+------------------------------------------------------------------+
